#include <cassert>
#include <cstring>
#include <iostream>
#include <cstdlib>   

#ifdef WIN32
#include <time.h>
#else
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <netinet/in.h>
#include <errno.h>
#endif

#include "udp.h"
#include "stun.h"

using namespace std;


void
usage()
{
   cerr << "Usage:" << endl
	<< "    ./client stunServerHostname [testNumber] [-v] [-p srcPort] [-i srcAddress] [-i2 secondSrcAddress]" << endl
	<< "For example, if the STUN server was larry.gloo.net, you could do:" << endl
	<< "    ./client larry.gloo.net" << endl
	<< "The testNumber is just used for special tests." << endl
	<< " test 1 runs test 1 from the RFC. For example:" << endl
	<< "    ./client larry.gloo.net 0" << endl << endl
        << "Return Values:" << endl
        << " -1  Generic Error" << endl << endl
        << "low order bits (mask 0x07)" << endl
        << "  0  No NAT Present (Open)" << endl
        << "  1  Full Cone NAT" << endl
        << "  2  Address Restricted Cone NAT" << endl
        << "  3  Port Restricted Cone NAT" << endl
        << "  4  Symmetric NAT" << endl
        << "  5  Symmetric Firewall" << endl
        << "  6  Blocked or Network Error" << endl << endl
        << "0x08 bit is set if the NAT supports hairpinning" << endl << endl;
}


int
main(int argc, char* argv[])
{
   assert( sizeof(UInt8 ) == 1 );
   assert( sizeof(UInt16) == 2 );
   assert( sizeof(UInt32) == 4 );
    
   initNetwork();
    
   clog << "STUN client version " 
        << STUN_VERSION_MAJOR << "." << STUN_VERSION_MINOR  << endl;
   
   int testNum = 0;
   bool verbose = false;
	
   StunAddress4 stunServerAddr;
   stunServerAddr.addr=0;
   int port=0;
   StunAddress4 sAddr;
   bool dualNic=false;
   StunAddress4 sAddr2;
   sAddr.addr=0; sAddr.port=0;
   int retval = 0;
   int retval1 = 0;
	
   for ( int arg = 1; arg<argc; arg++ )
   {
      if ( !strcmp( argv[arg] , "-v" ) )
      {
         verbose = true;
      }
      else if ( !strcmp( argv[arg] , "-i" ) )
      {
         arg++;
         if ( argc <= arg ) 
         {
            usage();
            exit(-1);
         }
         stunParseServerName(argv[arg], sAddr);
      }
      else if ( !strcmp( argv[arg] , "-i2" ) )
      {
        dualNic=true;
         arg++;
         if ( argc <= arg ) 
         {
            usage();
            exit(-1);
         }
         stunParseServerName(argv[arg], sAddr2);
      }
      else if ( !strcmp( argv[arg] , "-p" ) )
      {
         arg++;
         if ( argc <= arg ) 
         {
            usage();
            exit(-1);
         }
         port = strtol( argv[arg], NULL, 10);
      }
      else    
      {
        char* ptr;
        int t =  strtol( argv[arg], &ptr, 10 );
        if ( *ptr == 0 )
        { // conversion worked
           testNum = t;
           cerr << "running test number " << testNum  << endl; 
        }
        else
        {
           bool ret = stunParseServerName( argv[arg], stunServerAddr);
           if ( ret != true )
           {
              cerr << argv[arg] << " is not a valid host name " << endl;
              usage();
              exit(-1);
           }
	}	
      }
   }

   for ( int nic=0; nic<2; nic++ )
   {
      
      if (nic == 1 )
      {
	 retval1 = retval;
         if (!dualNic)
         {
            break;
         }
         sAddr = sAddr2;
      }
      
      sAddr.port=port;
      if ( stunServerAddr.addr == 0 )
      {
         usage();
         exit(-1);
      }
   
      if (testNum==0)
      {
         bool presPort=false;
         bool hairpin=false;
		
         NatType stype = stunNatType( stunServerAddr, verbose, &presPort, &hairpin, port, &sAddr);
		
         if ( nic == 0 )
         {
            if (dualNic)
            {
               clog << "Primary";
            }
            else
            {
               clog << "Internet";
            }
            clog << " connection is type: ";
         }
         else
         {
            clog << "Secondary connection is type: ";
         }
         
         switch (stype)
         {
            case StunTypeUnknown:
               clog << "Some error detetecting NAT type";
	       retval = 0x07;
               break;
            case StunTypeOpen:
               clog << "Open";
	       retval = 0; 
               break;
            case StunTypeConeNat:
               clog << "Full Cone Nat";
               if ( presPort ) clog << ", preserves ports"; else clog << ", random port";
               if ( hairpin  ) clog << ", will hairpin"; else clog << ", no hairpin";
               retval = 0x01;
               break;
            case StunTypeRestrictedNat:
               clog << "Address Restricted Nat";
               if ( presPort ) clog << ", preserves ports"; else clog << ", random port";
               if ( hairpin  ) clog << ", will hairpin"; else clog << ", no hairpin";
               retval = 0x02;
               break;
            case StunTypePortRestrictedNat:
               clog << "Port Restricted Nat";
               if ( presPort ) clog << ", preserves ports"; else clog << ", random port";
               if ( hairpin  ) clog << ", will hairpin"; else clog << ", no hairpin";
               retval = 0x03;
               break;
            case StunTypeSymNat:
               clog << "Symmetric Nat";
               if ( presPort ) clog << ", preserves ports"; else clog << ", random port";
               if ( hairpin  ) clog << ", will hairpin"; else clog << ", no hairpin";
               retval = 0x04;
               break;
            case StunTypeSymFirewall:
               clog << "Symmetric Firewall";
               if ( hairpin  ) clog << ", will hairpin"; else clog << ", no hairpin";
               retval = 0x05;
               break;
            case StunTypeBlocked:
               clog << "Blocked or could not reach STUN server";
               retval = 0x06;
               break;
            default:
               clog << stype;
               clog << "Unkown NAT type";
               retval = 0x07;  // Unknown NAT type
               break;
         }
         clog << endl;
         if (hairpin)
         {
             retval |= 0x08;
         }
      }
      else if (testNum==100)
      {
         Socket myFd = openPort(port,sAddr.addr);
      
         StunMessage req;
         memset(&req, 0, sizeof(StunMessage));
      
         StunAtrString username;
         StunAtrString password;
         username.sizeValue = 0;
         password.sizeValue = 0;
      
         stunBuildReqSimple( &req, username, 
                             false , false , 
                             0x0c );
      
         char buf[STUN_MAX_MESSAGE_SIZE];
         int len = STUN_MAX_MESSAGE_SIZE;
      
         len = stunEncodeMessage( req, buf, len, password,verbose );
      
         if ( verbose )
         {
            clog << "About to send msg of len " << len 
                 << " to " << stunServerAddr << endl;
         }
      
         while (1)
         {
            for ( int i=0; i<100; i++ )
            {
               sendMessage( myFd,
                            buf, len, 
                            stunServerAddr.addr, 
                            stunServerAddr.port );
            }
#ifdef WIN32 // !cj! TODO - should fix this up in windows
            clock_t now = clock();
            assert( CLOCKS_PER_SEC == 1000 );
            while ( clock() <= now+10 ) { };
#else
            usleep(10*1000);
#endif
         }
      }
      else if (testNum==-1)
      {
         StunAddress4 fd1Addr;
         int fd1 = stunOpenSocket( stunServerAddr, &fd1Addr );
        
         clog << "Opened a stun socket at " << fd1Addr << endl;
         StunAddress4 fd2Addr;
        
         int fd2 = stunOpenSocket( stunServerAddr, &fd2Addr );
         clog << "Opened a stun socket at " << fd2Addr << endl;
        
         StunAddress4 fd3Addr;
         int fd3,fd4;
         bool ok = stunOpenSocketPair(stunServerAddr,
                                      &fd3Addr,
                                      &fd3,
                                      &fd4 );
         if ( ok )
         {
            clog << "Opened a stun socket pair starting at " << fd3Addr << endl;
            closesocket(fd3);
            closesocket(fd4);
         }
         else
         {
            clog << "Opened a stun socket pair FAILED" << endl;
         }
        
         closesocket(fd1);
         closesocket(fd2);
      }
      else
      {
         stunTest( stunServerAddr,testNum,verbose,&sAddr );
      }
	
   }
   if (retval==-1 || retval1==-1) 
   {
	return -1;
   }
   if (retval == retval1) 
   {
	return retval;
   }
   else
   {
	return 0x07; // Unknown NAT type
   }
}


/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

// Local Variables:
// mode:c++
// c-file-style:"ellemtel"
// c-file-offsets:((case-label . +))
// indent-tabs-mode:nil
// End:
